#include <zephyr/kernel.h>
#include <zephyr/logging/log.h>
#include "sensor_dfu_upload_handler.h"
#include "sensor_dfu_handler.h"
#include <zephyr/storage/flash_map.h>
#include <zephyr/drivers/flash.h>
#include <flash_map_pm.h>
#include "image_hash_utils.h"

/* Note: if ever changing so that an image is uploaded from within the main gateway program it also need to be added so sensors are checked after the new image is uploaded.
         right now the need for update is only checkec when connecting relying on the fact that image upload is done externally to the main gateway program, and that it will
         reset the nrf when starting and thus all sensors will be connected again and the check kicks in*/
#define LOG_MODULE_NAME sensor_dfu_handler
LOG_MODULE_REGISTER(LOG_MODULE_NAME);

#define MAX_MSGS 12

#define OTA_TIMEOUT K_MINUTES(45)

typedef struct {
    int peer_index;
    ota_cb_t cb;
} ota_info_t;

K_MSGQ_DEFINE(ota_msgq, sizeof(ota_info_t), MAX_MSGS, 4);

#define DFU_STACK_SIZE 4096
#define DFU_PRIORITY 10

static void ota_thread(void *, void *, void *);
static void otatimer_timeout(struct k_timer *timer);

K_THREAD_DEFINE(ota_tid, DFU_STACK_SIZE,
                ota_thread, NULL, NULL, NULL,
                DFU_PRIORITY, 0, 0);

K_TIMER_DEFINE(ota_timer, otatimer_timeout, NULL);

atomic_t current_peer_index = ATOMIC_INIT(-1);

static void otatimer_timeout(struct k_timer *timer)
{
    int idx = atomic_get(&current_peer_index);

    LOG_INF("OTA timer expired");

    if (idx != -1) {
        dfu_operation_abort(idx);
    }
}

extern const uint8_t FooBin[] ;
extern const uint8_t FooBin_end[];

int8_t get_image_data(void** data, size_t* size, uint8_t hash[HASHSIZE])
{
    int8_t res = -1;
    size_t int_binsize = 0;
    size_t datasize = &(FooBin_end[0]) - &(FooBin[0]);
    uint8_t* datastart = &(FooBin[0]);

    LOG_INF("Data addr: %p", (void*)datastart);
	LOG_INF("Size %d", datasize);

    res = get_image_hash((void*)datastart, datasize, &int_binsize, hash);

    if (data != NULL) {
        *data = (void*)datastart;
    }

    if (size != NULL) {
        *size = int_binsize;
    }

    return res;
}

static void ota_device(int peer_index, ota_cb_t ota_cb)
{
    uint8_t imagehash[HASHSIZE];
    void* data = NULL;
    size_t size = 0;
    ota_status_t status = ota_status_undefined_error;
    int res = -1;

    atomic_set(&current_peer_index, peer_index);

    k_timer_start(&ota_timer, OTA_TIMEOUT, K_NO_WAIT);

    LOG_INF("Start OTA");
    res = get_image_data(&data, &size, imagehash);
    if (res == 0) {
        res = upload_image(peer_index, data, size, imagehash, HASHSIZE);
        if (res == 0) {
            res = send_smp_confirm(peer_index, imagehash, HASHSIZE);
            if (res == 0) {
                res = send_reset(peer_index);
                if (res == 0) {
                    LOG_INF("Update completed successfully");

                    status = ota_status_success;
                } else if (res == -2) {
                    LOG_ERR("Reset aborted");
                    status = reset_aborted;
                } else {
                    LOG_ERR("System reset failed");
                    status = ota_status_device_reset_failed;
                }
            } else if (res == -2) {
                LOG_ERR("Confirm aborted");
                status = confirm_aborted;
            } else {
                LOG_ERR("Image confirm failed");
                status = ota_status_image_confirm_failed;
            }
        } else if (res == -2) {
            LOG_ERR("Image upload aborted");
            status = image_upload_aborted;
        } else {
            LOG_ERR("Image upload failed");
            status = ota_status_image_upload_failed;
        }
    } else {
        LOG_ERR("Error getting image data");
        status = ota_status_get_image_data_failed;
    }

    k_timer_stop(&ota_timer);

    if (ota_cb != NULL) {
        ota_cb(peer_index, status);
    }

    atomic_set(&current_peer_index, -1);
}

static void ota_thread(void* unused1, void* unused2, void* unused3)
{
    while (1) {
        ota_info_t ota_info;

        if (k_msgq_get(&ota_msgq, &ota_info, K_FOREVER) == 0) {
            ota_device(ota_info.peer_index, ota_info.cb);
        } else {
            LOG_ERR("Error getting item from queue");
        }
    }
}

void peer_to_ota_thread(int index, ota_cb_t callback)
{
    ota_info_t ota_info = {.peer_index = index, .cb = callback};

    if (k_msgq_put(&ota_msgq, &ota_info, K_NO_WAIT) != 0) {
        callback(index, false);
    }
}

void ota_abort(int idx)
{
    if (idx == atomic_get(&current_peer_index)) {
        dfu_operation_abort(idx);
    }
}

