#include <zephyr/sys/util_macro.h>
#include <zephyr/mgmt/mcumgr/grp/img_mgmt/img_mgmt.h>
#include <zephyr/mgmt/mcumgr/grp/os_mgmt/os_mgmt.h>
#include <stdlib.h>
#include <zcbor_encode.h>
#include <zcbor_decode.h>
#include <zcbor_common.h>
#include <zephyr/drivers/flash.h>
#include <bluetooth/services/dfu_smp.h>
#include <zephyr/storage/flash_map.h>
#include <zephyr/logging/log.h>
#include "sensor_dfu_upload_handler.h"

#define LOG_MODULE_NAME therm_dfu_upload_handler
LOG_MODULE_REGISTER(LOG_MODULE_NAME);

#define EVENT_UPLOAD_IMAGE_RESPONSE 0x1
#define EVENT_CONFIRM_RESPONSE 0x2
#define EVENT_RESET_RESPONSE 0x4

#define RESPONSE_PROCESS_THREAD_STACK_SIZE 1024
#define RESPONSE_PROCESS_THREAD_PRIORITY 10

#define RESPONSE_PAYLOAD_SIZE 300

#define IMAGE_UPLOAD_OVERHEAD 95
#define MAX_SUPPORTED_CHUNKSIZE 200

#define CBOR_ENCODER_STATE_NUM 6
#define CBOR_DECODER_STATE_NUM 6

#define CBOR_MAP_MAX_ELEMENT_CNT 20

struct smp_buffer {
	struct bt_dfu_smp_header header;
	uint8_t payload[RESPONSE_PAYLOAD_SIZE];
};

struct context_common {
	struct smp_buffer smp_rsp_buff;
	struct smp_buffer smp_cmd;
	bool abort_flag;
};

struct sensor_dfu_context {
	struct context_common common;
	int32_t rc;
	int32_t offset;
	bool ok;
};
K_EVENT_DEFINE(dfu_response_finished);
static struct sensor_dfu_context context = {0};

K_EVENT_DEFINE(dfu_process_response);

static void reponse_process_thread(void *, void *, void *);

K_THREAD_DEFINE(response_process_tid, RESPONSE_PROCESS_THREAD_STACK_SIZE, reponse_process_thread, NULL, NULL, NULL, RESPONSE_PROCESS_THREAD_PRIORITY, 0, 0);

static int validate_header(uint16_t group_expected, uint8_t op_expected, uint8_t id_expected, void** payload, size_t* payload_len)
{
	int res = -1;
	uint16_t group = ((uint16_t)context.common.smp_rsp_buff.header.group_h8) << 8 | context.common.smp_rsp_buff.header.group_l8;

	if ((group == group_expected) && (context.common.smp_rsp_buff.header.op == op_expected) && (context.common.smp_rsp_buff.header.id == id_expected)) {
		*payload_len = (size_t)(((uint16_t)context.common.smp_rsp_buff.header.len_h8) << 8 | context.common.smp_rsp_buff.header.len_l8);
		*payload = context.common.smp_rsp_buff.payload;

		res = 0;
	} else {
		LOG_ERR("Unexpected value in header, group %d (expected %d) op %d (expected %d) id %d (expected %d)", group, group_expected, context.common.smp_rsp_buff.header.op, op_expected, context.common.smp_rsp_buff.header.id, id_expected);
	}

	return res;
}

static int decode_image(zcbor_state_t* zsd, uint8_t* hash)
{
	int res = -3;
	bool ok;
	struct zcbor_string value = {0};
	bool bootable = false;
	bool pending = false;
	bool permanent = false;
	bool dfu_slot_found = false;

	ok = zcbor_map_start_decode(zsd);
	if (ok) {
		do {
			int32_t slot_value = 0xffffffff;

			// Decoding image or slot key (depending on if image key is included or not)
			ok = zcbor_tstr_decode(zsd, &value);
			if (!ok) {
				LOG_ERR("Decoding error, image/slot key (err: %d)", zcbor_pop_error(zsd));
				res = -2;
				break;
			}

			// if image is included it must be decoded but if not the key should be slot
			if ((value.len == 5) && (strncmp(value.value, "image", 5) == 0)) {
				int32_t temp_val;

				ok = zcbor_int32_decode(zsd, &temp_val);
				if (!ok) {
					LOG_ERR("Decoding error, image value (err: %d)", zcbor_pop_error(zsd));
					res = -2;
					break;
				}

				//if image is included parse next key here (which should be slot, otherwise slot should already be in "value")
				ok = zcbor_tstr_decode(zsd, &value);
				if (!ok) {
					LOG_ERR("Decoding error, image/slot key (err: %d)", zcbor_pop_error(zsd));
					res = -2;
					break;
				}
			}

			if ((value.len != 4) || (strncmp(value.value, "slot", 4) != 0)) {
				LOG_ERR("unexpected value decoded, %d %.4s", value.len, value.value);
				res = -2;
				break;
			}

			// Decoding slot value
			ok = zcbor_int32_decode(zsd, &slot_value);
			if (!ok) {
				LOG_ERR("Decoding error, slot value (err: %d)", zcbor_pop_error(zsd));
				res = -2;
				break;
			}

			//"primary (running one) = 0 and secondary (for DFU dual-bank purposes) = 1"
			if (slot_value == 1) {
				LOG_INF("Found DFU slot");
				dfu_slot_found = true;
			}

			// Decoding version key
			ok = zcbor_tstr_decode(zsd, &value);
			if (!ok) {
				LOG_ERR("Decoding error, version key (err: %d)", zcbor_pop_error(zsd));
				res = -2;
				break;
			}

			// decoding version value
			ok = zcbor_tstr_decode(zsd, &value);
			if (!ok) {
				LOG_ERR("Decoding error, version value (err: %d)", zcbor_pop_error(zsd));
				res = -2;
				break;
			}

			// Decoding hash key
			ok = zcbor_tstr_decode(zsd, &value);
			if (!ok) {
				LOG_ERR("Decoding error, hash key (err: %d)", zcbor_pop_error(zsd));
				res = -2;
				break;
			}

			// decoding hash value
			ok = zcbor_bstr_decode(zsd, &value);
			if (!ok) {
				LOG_ERR("Decoding error, hash value (err: %d)", zcbor_pop_error(zsd));
				res = -2;
				break;
			}

			do {
				ok = zcbor_tstr_decode(zsd, &value);
				if (ok) {
					bool flag;

					ok = zcbor_bool_decode(zsd, &flag);
					if (ok) {
						//only if dfu slot
						if (slot_value == 1) {
							if ((value.len == 8) && (strncmp(value.value, "bootable", 8) == 0)) {
								bootable = flag;
							} else if ((value.len == 7) && (strncmp(value.value, "pending", 7) == 0)) {
								pending = flag;
							} else if ((value.len == 9) && (strncmp(value.value, "permanent", 9) == 0)) {
								permanent = flag;
							}
						}
					} else {
						LOG_ERR("Decoding error of value (but key found) (err: %d)", zcbor_pop_error(zsd));
						res = -2;
						break;
					}
				}else {
					LOG_ERR("Decoding error (no more values?), key (err: %d)", zcbor_pop_error(zsd));
					//This is probably not an error, just that there are no more values (and the number of values are not fixed)
				}
			} while (ok);
		} while(0);

		zcbor_map_end_decode(zsd);
	} else {
		LOG_INF("Non existing image or pasring error!");
	}

	// if res = -2 here there was an error parsing the response
	if (dfu_slot_found && bootable && pending && permanent) {
		LOG_INF("Image confirmed ok");
		res = 0;
	} else if (res != -2) {
		LOG_INF("Confirm repsonse parsed ok not dfu image or not in correct state");
		res = -1;
	} else {
		LOG_INF("Error occured while parsing image info");
	}

	return res;
}

static int decode_list_response(void* payload, size_t payload_len, uint8_t hash[HASHSIZE])
{
	int res = -1;

	zcbor_state_t zsd[10];
	struct zcbor_string value = {0};
	bool ok;

	zcbor_new_decode_state(zsd, ARRAY_SIZE(zsd), payload, payload_len, 5, NULL, 0);

	//zsd->constant_state->stop_on_error = false;

	ok = zcbor_map_start_decode(zsd);
	if (!ok) {
		LOG_ERR("Decoding error 1, start_decode (err: %d)", zcbor_pop_error(zsd));
		goto out;
	} 
		
	//Decoding images key
	ok = zcbor_tstr_decode(zsd, &value);
	if (!ok) {
		LOG_ERR("Decoding error 2, images key (err: %d)", zcbor_pop_error(zsd));
		goto mapout;
	}

	if ((value.len == 6) && (strncmp(value.value, "images", 6) == 0)) {
		ok = zcbor_list_start_decode(zsd);
		if (ok) {
			for (int image=0; image < 2; image++) {
				int decoderes = -1;

				LOG_INF("Decode image %d", image);
				decoderes = decode_image(zsd, hash);

				if (decoderes == 0) {
					res = 0;
					/*break parsing when correct info is found, this would probably make it impossible to resume
					 * parsing if not done, but that is not an issue the device should reboot soon
					 */
					break;
				} else if (decoderes != -1) {
					/* -1 is just not corect image found (and should continue) but other return codes indicate error and should abort */
					break;
				}
			}
		} else {
			LOG_ERR("Decoding error, start_decode images->list (err: %d)", zcbor_pop_error(zsd));
		}

		zcbor_list_end_decode(zsd);
	} else if ((value.len == 2) && (strncmp(value.value, "rc", 2) == 0)) {
		int32_t temp_val;

		ok = zcbor_int32_decode(zsd, &temp_val);
		if (ok) {
			LOG_ERR("Confirm responded with error %d", temp_val);
			res = -2;
		} else {
			LOG_ERR("Error decoding confirm rc value: %d", zcbor_pop_error(zsd));
			res = -3;
		}
	}

mapout:
	zcbor_map_end_decode(zsd);

out:
	return res;
}

static void process_upload_response(struct smp_buffer* smp_rsp_buff)
{
	size_t payload_len = 0;
	void* payload = NULL;

	if (validate_header(MGMT_GROUP_ID_IMAGE, MGMT_OP_WRITE_RSP, IMG_MGMT_ID_UPLOAD, &payload, &payload_len) == 0) {
		zcbor_state_t zsd[CBOR_DECODER_STATE_NUM];
		bool ok = false;

		zcbor_new_decode_state(zsd, ARRAY_SIZE(zsd), smp_rsp_buff->payload, payload_len, 1, NULL, 0);

		ok = zcbor_map_start_decode(zsd);
		if(ok) {
			do {
				int32_t temp_val;

				struct zcbor_string value = {0};

				//Decoding first key, this vill be rc in case of error and off if succesfull
				ok = zcbor_tstr_decode(zsd, &value);
				if (!ok) {
					LOG_ERR("Decoding error err: %d", zcbor_pop_error(zsd));
					break;
				}

				//Decoding first value... see above
				ok = zcbor_int32_decode(zsd, &temp_val);
				if (!ok) {
					LOG_ERR("Decoding error err: %d", zcbor_pop_error(zsd));
					break;
				}

				/* if error then rc value is encoded (offset not present/relevant) othervise
				 * offset is present but no rc value
				 */
				if ((value.len == 2) && (strncmp(value.value, "rc", 2) == 0)) {
					context.rc = temp_val;
					context.offset = 0;
				} else if ((value.len == 3) && (strncmp(value.value, "off", 3) == 0)) {
					context.rc = 0;
					context.offset = temp_val;
				}
			} while(0);

			zcbor_map_end_decode(zsd);
		} else {
			LOG_ERR("Decoding error, start_decode (err: %d)", zcbor_pop_error(zsd));
		}
	} else {
		LOG_ERR("Header validation failed");
	}
	k_event_set(&dfu_response_finished, 1);	
}

static void process_confirm_response(struct smp_buffer* smp_rsp_buff)
{
	size_t payload_len = 0;
	void* payload = NULL;

	if (validate_header(MGMT_GROUP_ID_IMAGE, MGMT_OP_WRITE_RSP, IMG_MGMT_ID_STATE, &payload, &payload_len) == 0) {
		uint8_t hash[HASHSIZE];
		int success = decode_list_response(payload, payload_len, hash);
		if (success == 0) {
			context.ok = true;
		}
	} else {
		LOG_ERR("Header validation failed");
	}

	k_event_set(&dfu_response_finished, 1);
}

static void process_reset_response(struct smp_buffer* smp_rsp_buff)
{
	size_t payload_len = 0;
	void* payload = NULL;

	if (validate_header(MGMT_GROUP_ID_OS, MGMT_OP_WRITE_RSP, OS_MGMT_ID_RESET, &payload, &payload_len) == 0) {
		bool ok = false;

		zcbor_state_t zsd[CBOR_DECODER_STATE_NUM];

		zcbor_new_decode_state(zsd, ARRAY_SIZE(zsd), smp_rsp_buff->payload, payload_len, 1, NULL, 0);

		ok = zcbor_map_start_decode(zsd);
		if (ok) {
			struct zcbor_string value = {0};

			//Decoding rc key, "only appears if non-zero (error condition).""
			ok = zcbor_tstr_decode(zsd, &value);
			if (ok) {
				if ((value.len == 2) || (strncmp(value.value, "rc", 2) == 0)) {
					int32_t rc_value;

					//Decoding rc value
					ok = zcbor_int32_decode(zsd, &rc_value);
					if (!ok) {
						LOG_ERR("Decoding error, rc value (err: %d)", zcbor_pop_error(zsd));
					} else {
						context.rc = rc_value;
					}
				} else {
					LOG_ERR("Invalid data received, expected key 'rc', got '%.2s' with len %d", value.value, value.len);
				}
			} else {
				//rc "may not appear if 0"
				context.rc = 0;
			}
		} else {
			LOG_ERR("Decoding error, start_decode (err: %d)", zcbor_pop_error(zsd));
		}

		zcbor_map_end_decode(zsd);
	} else {
		LOG_ERR("Header validation failed");
	}

	k_event_set(&dfu_response_finished, 1);
}

static void reponse_process_thread(void *, void *, void *)
{
	struct smp_buffer* smp_rsp_buff = &context.common.smp_rsp_buff;

	while (1) {
		uint32_t event = 0;

		event = k_event_wait(&dfu_process_response, EVENT_RESET_RESPONSE | EVENT_CONFIRM_RESPONSE | EVENT_UPLOAD_IMAGE_RESPONSE, false, K_FOREVER);
		k_event_clear(&dfu_process_response, event);

		switch (event) {
        	case EVENT_UPLOAD_IMAGE_RESPONSE: {
				process_upload_response(smp_rsp_buff);
            	break;
        	}

        	case EVENT_CONFIRM_RESPONSE: {
				process_confirm_response(smp_rsp_buff);
            	break;
        	}

        	case EVENT_RESET_RESPONSE: {
				process_reset_response(smp_rsp_buff);
            	break;
        	}

        	default: {
				LOG_ERR("Unknown process repsonse event");
            	break;
        	}
		}
	}
}

static void upload_response_process(struct bt_dfu_smp *dfu_smp)
{
	const struct bt_dfu_smp_rsp_state *rsp_state;
	struct smp_buffer* smp_rsp_buff = &context.common.smp_rsp_buff;
	uint8_t *p_outdata = (uint8_t *)smp_rsp_buff;

	context.rc = -1;

	rsp_state = bt_dfu_smp_rsp_state(dfu_smp);

	if (rsp_state->offset + rsp_state->chunk_size > sizeof(*smp_rsp_buff)) {
		LOG_ERR("Response size buffer overflow");
	} else {
		p_outdata += rsp_state->offset;
		memcpy(p_outdata,
		       rsp_state->data,
		       rsp_state->chunk_size);
	}

	if (bt_dfu_smp_rsp_total_check(dfu_smp)) {
		k_event_set(&dfu_process_response, EVENT_UPLOAD_IMAGE_RESPONSE);
	}
}

int upload_image(int peer_index, void* image_data, size_t image_size, uint8_t* hash, size_t hashsize)
{
	int res = -1;
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(peer_get_le_addr(peer_index), addr, BT_ADDR_LE_STR_LEN);
	memset(&context,0,sizeof(context));
	LOG_INF("upload_image %s size %d", addr, image_size);

	uint32_t curr_addr = 0;
	int chunk_size = 0;
	uint32_t last_print_offset = 0;
	
	context.common.abort_flag = false;
	context.rc = -1;
	context.offset = -1;

	chunk_size = bt_gatt_get_mtu(bt_dfu_smp_conn(get_peer_dfu_smp(peer_index))) - IMAGE_UPLOAD_OVERHEAD;

	if (chunk_size > MAX_SUPPORTED_CHUNKSIZE) {
		chunk_size = MAX_SUPPORTED_CHUNKSIZE;
	}

	LOG_INF("max chunksize %d", chunk_size);

	while (curr_addr < image_size) {
		int err = -1;
		int upload_chunk_size = chunk_size;
		zcbor_state_t zse[2];
		size_t payload_len = 0;
		uint8_t* data = (uint8_t*)image_data + curr_addr;

		zcbor_new_encode_state(zse, ARRAY_SIZE(zse), context.common.smp_cmd.payload, sizeof(context.common.smp_cmd.payload), 0);

		if(curr_addr + upload_chunk_size > image_size){
			upload_chunk_size = image_size - curr_addr;
		}

		if (((last_print_offset + 10240) < curr_addr) || (last_print_offset == 0)) {
			LOG_INF("Upload ongoing (%d / %d)", curr_addr, image_size);

			last_print_offset = curr_addr;
		}

		zcbor_map_start_encode(zse, CBOR_MAP_MAX_ELEMENT_CNT);
		zcbor_tstr_put_lit(zse, "image");
		zcbor_int64_put(zse, 0);
		zcbor_tstr_put_lit(zse, "data");
		zcbor_bstr_encode_ptr(zse, data, upload_chunk_size);
		zcbor_tstr_put_lit(zse, "len");
		zcbor_uint64_put(zse, image_size);
		zcbor_tstr_put_lit(zse, "off");
		zcbor_uint64_put(zse, curr_addr);
		zcbor_tstr_put_lit(zse, "sha");
		zcbor_bstr_encode_ptr(zse, hash, hashsize);
		zcbor_tstr_put_lit(zse, "upgrade");
		zcbor_bool_put(zse, false);
		zcbor_map_end_encode(zse, CBOR_MAP_MAX_ELEMENT_CNT);

		payload_len = (size_t)(zse->payload - context.common.smp_cmd.payload);

		context.common.smp_cmd.header.op = MGMT_OP_WRITE;
		context.common.smp_cmd.header.flags = 0;
		context.common.smp_cmd.header.len_h8 = (uint8_t)((payload_len >> 8) & 0xFF);
		context.common.smp_cmd.header.len_l8 = (uint8_t)((payload_len >> 0) & 0xFF);
		context.common.smp_cmd.header.group_h8 = 0;
		context.common.smp_cmd.header.group_l8 = MGMT_GROUP_ID_IMAGE;
		context.common.smp_cmd.header.seq = 0;
		context.common.smp_cmd.header.id = IMG_MGMT_ID_UPLOAD;
	
		k_event_set(&dfu_response_finished, 0);

		err = bt_dfu_smp_command(get_peer_dfu_smp(peer_index), upload_response_process,
					sizeof(context.common.smp_cmd.header) + payload_len,
					&(context.common.smp_cmd));
		if(err){
			LOG_ERR("bt_dfu_smp_command failed with %d", err);
			break;
		}

		//Wait until response is received until sending the next chunk
		k_event_wait(&dfu_response_finished, 1, false, K_FOREVER);

		if (context.common.abort_flag) {
			LOG_INF("Image upload operation aborted");
			context.rc = -2;
			res = -2;
			break;
		}

		curr_addr += upload_chunk_size;

		if (context.rc != 0) {
			LOG_ERR("Error during image upload %d", context.rc);
			break;
		}

		if (context.offset != curr_addr) {
			LOG_INF("Device repsonded with different offset %d (sent %d)", context.offset, curr_addr);

			if (context.offset < image_size) {
				LOG_INF("Moving");
				curr_addr = context.offset;
			}
		}
	}

	if (context.rc == 0) {
		LOG_INF("Upload done %s", addr);

		res = 0;
	} else {
		LOG_ERR("Error during image upload %s rc %d", addr, context.rc);
	}
	memset(&context,0,sizeof(context));

	return res;
}

static void smp_confirm_rsp_proc(struct bt_dfu_smp *dfu_smp)
{
	uint8_t *p_outdata = (uint8_t *)(&(context.common.smp_rsp_buff));
	const struct bt_dfu_smp_rsp_state *rsp_state;

	context.ok = false;

	rsp_state = bt_dfu_smp_rsp_state(dfu_smp);

	if (rsp_state->offset + rsp_state->chunk_size > sizeof(context.common.smp_rsp_buff)) {
		LOG_ERR("Response size buffer overflow");
	} else {
		p_outdata += rsp_state->offset;
		memcpy(p_outdata,
				rsp_state->data,
				rsp_state->chunk_size);
	}

	if (bt_dfu_smp_rsp_total_check(dfu_smp)) {
		k_event_set(&dfu_process_response, EVENT_CONFIRM_RESPONSE);
	}
}

int send_smp_confirm(int peer_index, uint8_t* hash, size_t hashsize)
{
	int res = -1;
	int err = -1;
	zcbor_state_t zse[2];
	size_t payload_len;
	char addr[BT_ADDR_LE_STR_LEN];

	context.ok = false;

	bt_addr_le_to_str(peer_get_le_addr(peer_index), addr, BT_ADDR_LE_STR_LEN);
	memset(&context,0,sizeof(context));
	LOG_INF("Confirm request %s", addr);
	
	context.common.abort_flag = false;
	context.ok = false;

	zcbor_new_encode_state(zse, ARRAY_SIZE(zse), context.common.smp_cmd.payload, sizeof(context.common.smp_cmd.payload), 0);

	zcbor_map_start_encode(zse, CBOR_MAP_MAX_ELEMENT_CNT);
	zcbor_tstr_put_lit(zse, "hash");
	zcbor_bstr_encode_ptr(zse, hash, hashsize);
	zcbor_tstr_put_lit(zse, "confirm");
	zcbor_bool_put(zse, true);

	zcbor_map_end_encode(zse, CBOR_MAP_MAX_ELEMENT_CNT);

	payload_len = (size_t)(zse->payload - context.common.smp_cmd.payload);

	context.common.smp_cmd.header.op = MGMT_OP_WRITE;
	context.common.smp_cmd.header.flags = 0;
	context.common.smp_cmd.header.len_h8 = (uint8_t)((payload_len >> 8) & 0xFF);
	context.common.smp_cmd.header.len_l8 = (uint8_t)((payload_len >> 0) & 0xFF);
	context.common.smp_cmd.header.group_h8 = 0;
	context.common.smp_cmd.header.group_l8 = MGMT_GROUP_ID_IMAGE;
	context.common.smp_cmd.header.seq = 0;
	context.common.smp_cmd.header.id = IMG_MGMT_ID_STATE;

	k_event_set(&dfu_response_finished, 0);

	// confirm has same response as list command
	err = bt_dfu_smp_command(get_peer_dfu_smp(peer_index), smp_confirm_rsp_proc,
			sizeof(context.common.smp_cmd.header) + payload_len,
			&(context.common.smp_cmd));

	if (err == 0) {
		//Wait until response is received
		k_event_wait(&dfu_response_finished, 1, false, K_FOREVER);

		if (context.common.abort_flag) {
			LOG_INF("Image confirm operation aborted");
			context.ok = false;
			res = -2;
		}

	} else {
		LOG_ERR("Error sending confirm command");
	}

	if (context.ok != true) {
		LOG_ERR("Error during confirm");
	} else {
		res = 0;
	}
	memset(&context,0,sizeof(context));
	return res;
}

static void smp_reset_rsp_proc(struct bt_dfu_smp *dfu_smp)
{
	const struct bt_dfu_smp_rsp_state *rsp_state;
	struct smp_buffer* smp_rsp_buff = &context.common.smp_rsp_buff;
	uint8_t *p_outdata = (uint8_t *)smp_rsp_buff;

	context.rc = -1;

	rsp_state = bt_dfu_smp_rsp_state(dfu_smp);

	if (rsp_state->offset + rsp_state->chunk_size > sizeof(*smp_rsp_buff)) {
		LOG_ERR("Response size buffer overflow");
	} else {
		p_outdata += rsp_state->offset;
		memcpy(p_outdata,
				rsp_state->data,
				rsp_state->chunk_size);
	}

	if (bt_dfu_smp_rsp_total_check(dfu_smp)) {
		k_event_set(&dfu_process_response, EVENT_RESET_RESPONSE);
	}
}

int send_reset(int peer_index)
{
	int res = -1;
	int err = -1;
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(peer_get_le_addr(peer_index), addr, BT_ADDR_LE_STR_LEN);
	memset(&context,0,sizeof(context));
	LOG_INF("Reset request %s", addr);
	
	context.common.abort_flag = false;

	context.common.smp_cmd.header.op = MGMT_OP_WRITE;
	context.common.smp_cmd.header.flags = 0;
	context.common.smp_cmd.header.len_h8 = 0;
	context.common.smp_cmd.header.len_l8 = 0;
	context.common.smp_cmd.header.group_h8 = 0;
	context.common.smp_cmd.header.group_l8 = MGMT_GROUP_ID_OS;
	context.common.smp_cmd.header.seq = 0;
	context.common.smp_cmd.header.id = OS_MGMT_ID_RESET;

	k_event_set(&dfu_response_finished, 0);

	err = bt_dfu_smp_command(get_peer_dfu_smp(peer_index), smp_reset_rsp_proc,
			sizeof(context.common.smp_cmd.header),
			&(context.common.smp_cmd));

	if (err == 0) {
		//Wait until response is received
		k_event_wait(&dfu_response_finished, 1, false, K_FOREVER);

		if (context.common.abort_flag) {
			LOG_INF("Reset operation aborted");
			context.rc = -2;
		}

	} else {
		LOG_ERR("Error sending reset command");
	}

	if (context.rc != 0) {
		LOG_ERR("Error during reset, rc %d", context.rc);
	}

	res = context.rc;
	memset(&context,0,sizeof(context));
	return res;
}

void dfu_operation_abort(int peer_index)
{
	LOG_ERR("Abort current dfu operation");
	context.common.abort_flag = true;
	k_event_set(&dfu_response_finished, 1);
}